open Client

module LexUtil = MenhirLib.LexerUtil

(* -------------------------------------------------------------------------- *)

(* A wrapper over the client's [translate] function. We consider ill-typedness
   as normal, since our terms are randomly generated, so we translate the client
   exceptions to [None]. *)

let print_type ty =
  PPrint.(ToChannel.pretty 0.9 80 stdout (FPrinter.print_type ty ^^ hardline))

let translate e t =
  try
    Some (Infer.translate e t)
  with
  | Infer.Unbound (range, s) ->
      Printf.eprintf "%!%sType error: unbound variable \"%s\".\n"
        (LexUtil.range range) s;
      None
  | Infer.Cycle (_range, ty) ->
      if Config.verbose then begin
        Printf.fprintf stdout "Type error: a cyclic type arose.\n";
        print_type ty
      end;
      None
  | Infer.Unify (range, ty1, ty2) ->
      if Config.verbose then begin
        Printf.eprintf "%!%sType error: type mismatch.\n"
            (LexUtil.range range);
        Printf.fprintf stdout "Type error: mismatch between the type:\n";
        print_type ty1;
        Printf.fprintf stdout "and the type:\n";
        print_type ty2
      end;
      None

let equal_term t1 t2 =
  ML.compare_term t1 t2 = 0


(* -------------------------------------------------------------------------- *)

let wrap parser lexbuf =
  let lexbuf = LexUtil.init "test" lexbuf in
  try parser MLLexer.read lexbuf
  with exn ->
    let range = Lexing.(lexbuf.lex_start_p, lexbuf.lex_curr_p) in
    Printf.eprintf "%!%sSyntax error.\n%!"
      (LexUtil.range range);
    raise exn

(* Main parsing function *)
let term lexbuf = wrap MLParser.self_contained_term lexbuf
let program lexbuf = wrap MLParser.prog lexbuf
let type_decls lexbuf = wrap MLParser.self_contained_type_decls lexbuf

exception ParsingError

let ast_from_string s =
  let lexbuf = Lexing.from_string s in
  try program lexbuf
  with _ -> raise ParsingError

let letify xts =
  let rec aux xts =
    match xts with
    | [] ->
        assert false
    | [(_, t)] ->
        t
    | (x, t) :: xts ->
        ML.Let (ML.dummy_pos, x, t, aux xts)
  in
  aux xts

let open_uses filenames =
  let on_file typedecl_env filename =
    let ch = open_in filename in
    let lexbuf = Lexing.from_channel ch in
    let typedecl_env' = type_decls lexbuf in
    close_in ch;
    Datatype.Env.union typedecl_env typedecl_env'
  in
  List.fold_left on_file Datatype.Env.empty filenames

let from_string typedecl s =
  let s = typedecl ^ "\nlet _ = " ^ s in
  let (uses, typedecl_env, xts) = ast_from_string s in
  let typedecl_env' = open_uses uses in
  let ast = letify xts in
  (Datatype.Env.union typedecl_env typedecl_env', ast)

let from_file filename =
  let ch = open_in filename in
  let lexbuf = Lexing.from_channel ch in
  let (uses, typedecl_env, xts) = program lexbuf in
  close_in ch;
  let typedecl_env' = open_uses uses in
  let ast = letify xts in
  (Datatype.Env.union typedecl_env typedecl_env', ast)

(* -------------------------------------------------------------------------- *)

(* Running all passes over a single ML term. *)

let test_with_args
  ~(success:int ref) ~(total:int ref) (env : ML.datatype_env) (t : ML.term)
  : bool
  =
  incr total;
  (* We print the term, parse it and compare with the given term *)
  let s = MLPrinter.to_string t in
  let lexbuf = Lexing.from_string s in
  let t' = MLParser.self_contained_term MLLexer.read lexbuf in
  assert (equal_term t t');
  let outcome_env =
    ML2F.translate_env env in
  print_endline "Type inference and translation to System F...";
  let outcome =
    translate env t
  in
  match outcome with
  | None ->
      (* This term is ill-typed. This is considered a normal outcome, since
         our terms can be randomly generated. *)
      true
  | Some (t : F.nominal_term) ->
      CheckF.test outcome_env t;
      (* Everything seems to be OK. *)
      incr success;
      true

let test env t =
  test_with_args ~success:(ref 0) ~total:(ref 0) env t
