(******************************************************************************)
(*                                                                            *)
(*                                  Inferno                                   *)
(*                                                                            *)
(*                       François Pottier, Inria Paris                        *)
(*                                                                            *)
(*  Copyright Inria. All rights reserved. This file is distributed under the  *)
(*  terms of the MIT License, as described in the file LICENSE.               *)
(*                                                                            *)
(******************************************************************************)

open UnifierSig
open SolverSig

module Make
  (X : TEVAR)
  (S : STRUCTURE)
  (O : OUTPUT with type 'a structure = 'a S.structure)
: sig

  (* The type [tevar] of term variables is provided by [X]. *)
  open X

  (* The type ['a structure] of shallow types is provided by [S]
     (and repeated by [O]). *)

  (* The types [tyvar] and [ty] of decoded type variables and decoded types
     are provided by the module [O]. *)
  open O

  (**The type [variable] describes the solver's type variables. *)
  type variable

  (* ---------------------------------------------------------------------- *)

  (* The type ['a co] is the type of a constraint, which, once solved, produces
     a result of type ['a]. Put another way, it is the type of a computation
     which, once run to completion, produces a result of type ['a]. *)
  type 'a co

  (* ---------------------------------------------------------------------- *)

  (* The applicative functor interface: tautology, conjunction, post-processing. *)

  (* The following three operations form the applicative functor interface.
     [pure] represents the true constraint, while [(and+)] builds a conjunction
     of constraints. [(let+)] does not per se produce any constraint; it only
     performs postprocessing, i.e., it applies a user-supplied function to
     the value that is obtained after constraint solving. *)
  val pure: 'a -> 'a co

  val (let+): 'a co -> ('a -> 'b) -> 'b co
  val (and+): 'a co -> 'b co -> ('a * 'b) co

  (* ---------------------------------------------------------------------- *)

  (* Equations. *)

  (* [v -- w] is an equality constraint on the type variables [v] and [w]. *)
  val (--): variable -> variable -> unit co

  (* [v --- t] is analogous, but its right-hand side is a shallow type, as
     opposed to a type variable. In fact, [---] is just [lift (--)]. *)
  val (---): variable -> variable structure -> unit co

  (* ---------------------------------------------------------------------- *)

  (* Existential quantification. *)

  (* A binder of type [('a, 'r) binder] introduces a value of type ['a] during
     the construction of a constraint of type ['r co]. For example, the binder
     [exist], documented below, introduces a fresh type variable [v]. A binder
     can be used in two (equivalent) styles: in indirect style, one can write
     [exist (fun v -> ...)]; in direct style, one writes [let@ v = exist in
     ...]. *)
  type ('a, 'r) binder =
    ('a -> 'r co) -> 'r co

  val (let@) : ('a, 'r) binder -> ('a -> 'r co) -> 'r co

  (* The binder [exist] introduces a fresh existentially quantified type
     variable [v]. *)
  val exist:                            (variable, 'r) binder

  (* [construct t c] is analogous to [exist c], but additionally constrains
     the type variable [v] to be equal to the type [t]. So, it is really a
     way of constructing a variable that stands for a shallow type. *)
  val construct:  variable structure -> (variable, 'r) binder

  (* [witness v] can be thought of as the witness (that is, the final value)
     of the (existentially quantified) type variable [v]. It is a decoded
     type. Of course, because the final value of [v] cannot be known until
     the constraint construction and constraint solving phases are over,
     [witness v] cannot simply have type [ty]; instead, it has type [ty co],
     that is, it is a promise for a decoded type, or a deferred decoded type. *)
  val witness: variable -> ty co

  (* ---------------------------------------------------------------------- *)

  (* Deep types. *)
  type deep_ty =
    | DeepVar of variable
    | DeepStructure of deep_ty S.structure

  (* [build dty c] constructs a variable [v] that stands for the deep type
     [dty] and passes [v] to the function [c]. *)
  val build: deep_ty -> (variable, 'r) binder

  (* ---------------------------------------------------------------------- *)

  (* A utility. *)

  (* If [f] is a binary predicate whose second argument is a type variable,
     then [lift f] is a binary predicate whose second argument is a shallow
     type. It is defined in terms of [construct_] and [f]. *)
  val lift: ('a -> variable -> 'b co) -> 'a -> variable structure -> 'b co

  (* ---------------------------------------------------------------------- *)

  (* Application of constraint abstractions, a.k.a. instantiation. *)

  (* The constraint [instance x v] requires [v] to be an instance of the type
     scheme denoted by the term variable [x]. Or, in other words, it is an
     application of the constraint abstraction denoted by [x] to the variable
     [v]. *)
  val instance: tevar -> variable -> ty list co

  (* [instance_ x v] is equivalent to [instance x v <$$> ignore]. *)
  val instance_: tevar -> variable -> unit co

  (* ---------------------------------------------------------------------- *)

  (* Construction of constraint abstractions, a.k.a. generalization. *)

  (* [def x v c] binds the term variable [x] to the trivial (monomorphic) type
     scheme [v] in the constraint [c]. *)
  val def: tevar -> variable -> 'a co -> 'a co

  (* [let1 x c1 c2] binds the term variable [x] to the constraint abstraction
     [c1] in the constraint [c2]. (Technically, [c1] is a function of a fresh
     type variable to a constraint, as in [exist].) The resulting constraint
     produces a tuple of 4 results, namely:
     - the (decoded) type scheme that was assigned to [x] while solving [c2].
     - a list [vs] of (decoded) type variables that may appear in [a1], and (hence)
       should be universally bound by the client in [a1]. In a typical scenario,
       [a1] is a System F term, and the client will build the type abstraction
       [\Lambda vs.a1].
     - the value [a1] produced by the constraint [c1].
     - the value [a2] produced by the constraint [c2]. *)
  val let1: tevar -> (variable -> 'a co) -> 'b co ->
            (scheme * tyvar list * 'a * 'b) co

  (* [let0 c] has the same meaning as [c], but, like [let1], produces a list [vs]
     of the type variables that may appear in the result of [c]. The argument of
     [run] should always be an application of [let0] -- see below. *)
  val let0: 'a co  -> (tyvar list * 'a) co

  (* [letn xs c1 c2] binds [n] term variables [xs] to [n] constraint abstractions
     in the constraint [c2]. Here, [c1] is a function of [n] fresh type variables
     [vs] to a constraint. The [i]-th term variable, [x_i], ends up bound to the
     constraint abstraction of the [i]-th type variable in [c_1], which one could
     write [\lambda v_i.c_1]. *)
  val letn: tevar list -> (variable list -> 'a co) -> 'b co ->
            (scheme list * tyvar list * 'a * 'b) co

  (* ---------------------------------------------------------------------- *)

  (* Correlation with the source code. *)

  (**The type [range] describes a range in the source code. It is not
     interpreted by the solver. It is used only as part of error reports. *)
  type range =
    Lexing.position * Lexing.position

  (* The constraint [correlate range c] is equivalent to [c], but records
     that this constraint is correlated with the range [range] in the
     source code. This information is used in error reports. *)
  val correlate: range -> 'a co -> 'a co

  val pprint : _ co -> PPrint.document
  val print : Format.formatter -> _ co -> unit

  (* The function [solve] is parameterized by the flag [rectypes], which
     indicates whether recursive types are permitted. It expects a constraint
     and solves it; that is, either it fails with an exception, or it succeeds
     and returns a mapping from constraint variables to inferred types. *)

  exception Unbound of range * tevar
  exception Unify of range * O.ty * O.ty
  exception Cycle of range * O.ty

  val solve: rectypes:bool -> 'a co -> 'a

end
